/*
MODEL.H file for the FTERMHYBRID model.
Written by Gary Felder (gfelder@physics.stanford.edu)
Last Modified 3/7/01

This file contains the model specific functions and definitions for the model:
V = lambda [|phi|^2 |sigma|^2 + |phi|^2 |sigmabar|^2 + |sigma|^2 |sigmabar|^2 - v^2 (sigma_R sigmabar_R - sigma_I sigmabar_I) + 1/4 v^4]
in which phi, sigma, and sigmabar are complex fields with real and imaginary parts f_R and f_I.
Comments: This potential comes from the superpotential W=sqrt(lambda)/2 Phi (4 Sigmabar Sigma - v^2). This is a hybrid model in which slow roll inflation is driven by the field phi, but terminated by symmetry breaking of the fields sigma and sigmabar. The potential should include a slow roll term such as 1/2 m^2 phi^2 but this has been left out because it should be negligible during reheating. This model is currently only implemented for use without expansion.

We do not use the default rescalings (see below), but rather set A=sqrt(2)/v, B=sqrt(lambda) v
In the absence of expansion the scale v is completely eliminated from both the equations of motion and the initial fluctuations. However it can still turn up, for example, in setting the initial velocity of the phi field.
The potential in program units is given by V_{pr} = 1/2 |phi_pr|^2 |sigma_pr|^2 + 1/2 |phi_pr|^2 |sigmabar_pr|^2 + 1/2 |sigma_pr|^2 |sigmabar_pr|^2 - (sigma_{R,pr} sigmabar_{R,pr} - sigma_{I,pr} sigmabar_{I,pr}) + 1/2
*/

/*
General comments about the model.h file:
This file contains the following functions - all called externally
modelinfo(FILE *info_) outputs information about the model and model-specific parameters to a file.
modelinitialize() performs any model-specific initialization
potential_energy(int term, float *field_values) calculates the average potential energy density term by term. The variable num_potential_terms (just above this function) specifies how many separate potential terms are used in this model.
dvdf(int fld, int i, int j, int k) calculates the potential term in the equation of motion, dV/dfield, for the field fld at the lattice point (i,j,k)
effective_mass(float mass_sq[], float *field_values) calculates the square masses of the fields and puts them in the array mass_sq. The parameter beginning tells the function to use initial field values - if this parameter is zero then the field quantities will be calculated dynamically.
model_output(int flush, char *ext_) allows each model to include its own specialized output function(s). The parameter flush is set to 1 when infrequent calculations are being performed and 0 otherwise. The string ext_ gives the extension for output filenames.
*/

/* - This section should be copied into parameters.h when using this model
// ---Adjustable parameters for FTERMFLDHYBRID model--- //
const float v=1.e-3; // In the absence of expansion this factor has no effect
const float lambda=1.e-2; // Self-coupling of inflaton
const float vinit=-1.13e+4*sqrt(lambda)*pw2(v); // Initial velocity of the phi field - Should be entered in initderivs in parameters.h
// The following options control output functions for the F-Term model
const int sstrings=1; const float tstrings=20.; const float string_cutoff=.01; // Find "strings", i.e. points with |sigma|^2<string_cutoff. Start when t>=tstrings
const int srmshistograms=1; // Generate histograms of rms values of phi, sigma, and sigmabar
const int srphimeansvars=1; // Output means and variances of phases and amplitudes
const int srmshistograms2d=1; // Generate 2d histograms of rms values of phi and sigma
*/

// Rescaling parameters.
// The program variables ("_pr") are defined as
//   f_pr = rescale_A a^rescale_r f (f=field value)
//   x_pr = rescale_B x (x=distance)
//   dt_pr = rescale_B a^rescale_s dt (t=time)
// For this model A and B are not set using the defaults. The rescalings r and s are irrelevant without expansion.
const float rescale_A=sqrt(2.)/v;
const float rescale_B=sqrt(lambda)*v;
const float rescale_r=0.;
// The value of S in terms of R SHOULD NOT be changed.
const float rescale_s=2.*rescale_r-3.;

// Other global variables
// The array model_vars is intended to hold any model-specific, non-constant global variables.
// These variables should be initialized in modelinitialize() below
// Even if you're not using any, num_model_vars should be at least 1 to keep some compilers happy.
const int num_model_vars=1;
// Model specific variables: None are defined for this model.
extern float model_vars[num_model_vars];

// Macros to make the equations more readable: The values of fld are 0=Phi_R,1=Phi_I,2=Sigma_R,3=Sigma_I,4=Sigmabar_R,5=Sigmabar_I
#define PHIR FIELD(0)
#define PHII FIELD(1)
#define SIGMAR FIELD(2)
#define SIGMAI FIELD(3)
#define SIGMABARR FIELD(4)
#define SIGMABARI FIELD(5)
// Square magnitudes of the complex fields
#define PHI2 (pw2(PHIR)+pw2(PHII))
#define SIGMA2 (pw2(SIGMAR)+pw2(SIGMAI))
#define SIGMABAR2 (pw2(SIGMABARR)+pw2(SIGMABARI))

// Model specific details about the run to be output to an information file
inline void modelinfo(FILE *info_)
{
  // Name and description of model
  fprintf(info_,"F-Term Hybrid Model\n");
  fprintf(info_,"V = lambda [|phi|^2 |sigma|^2 + |phi|^2 |sigmabar|^2 + |sigma|^2 |sigmabar|^2 - v^2 (sigma_R sigmabar_R - sigma_I sigmabar_I) + 1/4 v^4]\n\n");

  // Model specific parameter values
  fprintf(info_,"lambda = %e\n",lambda);
  fprintf(info_,"v = %e\n",v);
}

// Perform any model specific initialization
// This function is called twice, once before initializing the fields (which_call=1) and once after (which_call=2)
inline void modelinitialize(int which_call)
{
  if(which_call==1)
  {
    if(nflds!=6)
    {
      printf("Number of fields for FTERMHYBRID model must be 6. Exiting.\n");
      exit(1);
    }
    if(expansion!=0)
    {
      printf("FTERMHYBRID model has not been implemented with expansion. To run  this model reset expansion to zero.\n");
      exit(1);
    }
  }
}

// The constant num_potential_terms must be defined for use by outside functions
// Terms: term=0: lambda |phi|^2 (|sigma|^2 + |sigmabar|^2) --- term=1: 1/4 lambda |2 sigmabar sigma - v^2|^2
const int num_potential_terms=2; // Number of terms that are calculated separately in the potential
// Potential energy terms
// See documentation for normalization of these terms.
// When setting initial conditions field values will be supplied in the array field_values. Otherwise this function will calculate them on the lattice.
inline float potential_energy(int term, float *field_values)
{
  DECLARE_INDICES
  float potential=0.;

  if(field_values==NULL) // If no values are given calculate averages on the lattice
  {
    // Loop over grid to calculate potential term
    LOOP
    {
      if(term==0)
        potential += PHI2*(SIGMA2+SIGMABAR2);
      else if(term==1)
        potential += .5*SIGMA2*SIGMABAR2-SIGMAR*SIGMABARR+SIGMAI*SIGMABARI;
    }

    // Convert sum to average
    potential /= gridsize;
  }
  else // If field values are given then use them instead
  {
    if(term==0)
      potential = (pw2(field_values[0])+pw2(field_values[1]))*(pw2(field_values[2])+pw2(field_values[3])+pw2(field_values[4])+pw2(field_values[5]));
    else if(term==1)
      potential = .5*(pw2(field_values[2])+pw2(field_values[3]))*(pw2(field_values[4])+pw2(field_values[5]))-field_values[2]*field_values[4]+field_values[3]*field_values[5];
  }

  // Include numerical coefficients
  if(term==0)
    potential *= .5;
  else if(term==1) // Constant term was left out above to be added here
    potential += .5;

  return (potential);
}

// Potential terms in the equations of motion, dV/dfield, evaluated at point (i,j,k)
// See documentation for details on the normalization of these terms
inline float dvdf(int fld, INDEXLIST)
{
  // Note: This function could probably be sped up for this model by calculating the square magnitudes of all the fields only once at each lattice point. Perhaps I should implement this in the future.
  if(fld==0) // Phi_R
    return( (SIGMA2+SIGMABAR2)*PHIR );
  else if(fld==1) // Phi_I
    return( (SIGMA2+SIGMABAR2)*PHII );
  else if(fld==2) // Sigma_R
    return( (PHI2+SIGMABAR2)*SIGMAR - SIGMABARR);
  else if(fld==3) // Sigma_I
    return( (PHI2+SIGMABAR2)*SIGMAI + SIGMABARI);
  else if(fld==4) // Sigmabar_R
    return( (PHI2+SIGMA2)*SIGMABARR - SIGMAR);
  else // Sigmabar_I
    return( (PHI2+SIGMA2)*SIGMABARI + SIGMAI);
}

// Calculate effective mass squared and put it into the array mass_sq[] (used for initial conditions and power spectra)
// See documentation for normalization of these terms
// When setting initial conditions field values will be supplied in the array field_values. Otherwise this function will calculate them on the lattice.
inline void effective_mass(float mass_sq[], float *field_values)
{
  DECLARE_INDICES
  int fld;
  float fldsqrd[nflds]; // Square value of field
  float correction; // Used to adjust masses by the appropriate power of the scale factor. (See documentation.)

  // Loop over fields to find mean-square value
  if(field_values==NULL) // If no values are given calculate averages on the lattice
  {
    for(fld=0;fld<nflds;fld++)
    {
      fldsqrd[fld]=0.;
      LOOP
        fldsqrd[fld]+=pw2(FIELD(fld));
      fldsqrd[fld] /= (float)gridsize;
    }
  }
  else // If field values are given then use them instead
    for(fld=0;fld<nflds;fld++)
      fldsqrd[fld]=0.; // This sets initial masses to zero. Technically the mass matrix isn't diagonal, so this is something of a kludge.

  mass_sq[0] = fldsqrd[2] + fldsqrd[3] + fldsqrd[4] + fldsqrd[5];
  mass_sq[1] = mass_sq[0];
  mass_sq[2] = fldsqrd[0] + fldsqrd[1] + fldsqrd[4] + fldsqrd[5];
  mass_sq[3] = mass_sq[2];
  mass_sq[4] = fldsqrd[0] + fldsqrd[1] + fldsqrd[2] + fldsqrd[3];
  mass_sq[5] = mass_sq[4];

  // Put in scale factor correction. This calculation should be the same for all models.
  if(expansion>0) // If there's no expansion don't bother with this.
  {
    correction = pow(a,2.*rescale_s+2.);
    for(fld=0;fld<nflds;fld++)
      mass_sq[fld] *= correction;
  }
}

////////////////////////////////////////////////////////////////////////////////////
// Model-specific output functions - The function model_output is called externally/
////////////////////////////////////////////////////////////////////////////////////

// This function finds all points on the lattice in "strings," i.e. with magnitude(sigma) less than a preset cutoff
inline void find_strings(char *ext_)
{
  static FILE *strings_,*stringtimes_;
  char name_[500]; // Used for filenames
  int i,j,k;
  int numpts=0; // Number of points in strings at a given time
  float mag2; // Square magnitude of sigma at a given point

  static int first=1;
  if(first) // Open output files
  {
    sprintf(name_,"strings%s",ext_);
    strings_=fopen(name_,mode_);
    sprintf(name_,"stringtimes%s",ext_);
    stringtimes_=fopen(name_,mode_);
    first=0;
  }

  LOOP
  {
    mag2 = pw2(FIELD(2))+pw2(FIELD(3));
    if(mag2<string_cutoff) // All values below the cutoff are considered part of a string
    {
      if(NDIMS==1)
        fprintf(strings_,"%d\n",i);
      else if(NDIMS==2)
        fprintf(strings_,"%d %d\n",i,j);
      else if(NDIMS==3)
        fprintf(strings_,"%d %d %d\n",i,j,k);
      numpts++;
    }
  }
  fprintf(stringtimes_,"%f %d\n",t,numpts);
  fflush(strings_);
  fflush(stringtimes_);
  
  return;
}

// This function outputs histograms of the rms values of all complex fields
// This array rmshistflds lists, pairwise, the real and imaginary parts of each complex field. It must have an even number of entries
const int rmshistflds[]={0,1,2,3,4,5};
inline void rms_histograms(char *ext_)
{
  static FILE **rmshist_,*rmshisttimes_;
  char name_[500]; // Used for filenames
  int i,j,k,pair; // pair iterates over pairs of fields, each of which consists of f_R and f_I
  int fld1,fld2; // Indices of the fields for a given histogram
  int binnum; // Index of bin for a given field value
  float binfreq[nbins]; // The frequency of field values occurring within each bin
  float bmin,bmax,df; // Minimum and maximum field values for each field and spacing (in field values) between bins
  float rms; // RMS value of the complex field. Using this variable avoids recalculating it several times at each point.

  static int first=1,numhists;
  if(first) // Open output files
  {
    numhists = (sizeof(rmshistflds)/sizeof(rmshistflds[0])/2); // Number of pairs in field list
    rmshist_ = new FILE *[numhists]; // Allocate an array of FILE pointers
    for(pair=0;pair<numhists;pair++)
    {
      sprintf(name_,"rmshist%d%s",pair,ext_);
      rmshist_[pair]=fopen(name_,mode_);
    }
    sprintf(name_,"rmshisttimes%s",ext_);
    rmshisttimes_=fopen(name_,mode_);
    first=0;
  }

  fprintf(rmshisttimes_,"%f",t); // Output time at which power spectra were recorded
  
  for(pair=0;pair<numhists;pair++)
  {
    i=0;j=0;k=0;
    fld1=2*pair;
    fld2=2*pair+1;
    // Find the minimum and maximum rms values
    bmin=sqrt(pw2(FIELD(fld1))+pw2(FIELD(fld2)));
    bmax=bmin;
    LOOP
    {
      rms=sqrt(pw2(FIELD(fld1))+pw2(FIELD(fld2))); // Calculate RMS value
      bmin = (rms<bmin ? rms : bmin);
      bmax = (rms>bmax ? rms : bmax);
    }
  
    // Find the difference (in field value) between successive bins
    df=(bmax-bmin)/(float)(nbins); // bmin will be at the bottom of the first bin and bmax at the top of the last
  
    // Initialize all frequencies to zero
    for(i=0;i<nbins;i++)
      binfreq[i]=0.;
  
    // Iterate over grid to determine bin frequencies
    LOOP
    {
      rms=sqrt(pw2(FIELD(fld1))+pw2(FIELD(fld2))); // Calculate RMS value
      binnum=(int)((rms-bmin)/df); // Find index of bin for each value
      if(rms>=bmax) // The maximal field value is at the top of the highest bin
        binnum=nbins-1;
      if(binnum>=0 && binnum<nbins) // Increment frequency in the appropriate bin
        binfreq[binnum]++;
      else // This should never happen, but it doesn't hurt to have an extra check.
      {
        printf("Error in rmshist routine. Invalid bin number.\n");
        fprintf(rmshist_[pair],"Invalid bin number\n");
      }
    } // End of loop over grid
  
    // Output results
    for(i=0;i<nbins;i++)
      fprintf(rmshist_[pair],"%e\n",binfreq[i]/(float)gridsize); // Output bin frequency normalized so the total equals 1
    fprintf(rmshist_[pair],"\n"); // Stick a blank line between times to make the file more readable
    fflush(rmshist_[pair]);
    fprintf(rmshisttimes_," %e %e",bmin,df); // Output the starting point and stepsize for the bins at each time
  } // End of loop over pairs
  fprintf(rmshisttimes_,"\n");
  fflush(rmshisttimes_);

  return;
}

// This function computes means and variances of the phase and amplitude of each complex field
// This array rphiflds lists, pairwise, the real and imaginary parts of each complex field. It must have an even number of entries
const int rphiflds[]={0,1,2,3,4,5};
inline void rphimeansvars(char *ext_, int flush)
{
  static FILE *rphimeans_,*rphivars_;
  char name_[500]; // Used for filenames
  DECLARE_INDICES
  int pair; // pair iterates over pairs of fields, each of which consists of f_R and f_I
  int fld1,fld2; // Indices of the fields for a given histogram
  double avr,varr,avphi,varphi,avr_sq,avphi_sq;
  float fr,fi;

  static int first=1,numfields; // Numfields is the number of COMPLEX fields
  if(first) // Open output files
  {
    numfields = (sizeof(rphiflds)/sizeof(rphiflds[0])/2); // Number of pairs in field list
    sprintf(name_,"rphimeans%s",ext_);
    rphimeans_=fopen(name_,mode_);
    sprintf(name_,"rphivars%s",ext_);
    rphivars_=fopen(name_,mode_);
    first=0;
  }

  fprintf(rphimeans_,"%f",t);
  fprintf(rphivars_,"%f",t);
  for(pair=0;pair<numfields;pair++)
  {
    fld1=2*pair;
    fld2=2*pair+1;
    avr=0.;
    varr=0.;
    avphi=0.;
    varphi=0.;
    LOOP
    {
      fr = FIELD(fld1);
      fi = FIELD(fld2);
      avr += sqrt(pw2(fr) + pw2(fi));
      avphi += atan2(fi,fr);
    }
    avr = avr/(double)gridsize;
    avphi = avphi/(double)gridsize;
    avr_sq = pw2(avr);
    avphi_sq = pw2(avphi);
    LOOP
    {
      fr = FIELD(fld1);
      fi = FIELD(fld2);
      varr += pw2(fr) + pw2(fi) - avr_sq;
      varphi += pw2(atan2(fi,fr)) - avphi_sq;
    }
    varr = varr/(double)gridsize;
    varphi = varphi/(double)gridsize;
    fprintf(rphimeans_," %e %e",avr,avphi);
    fprintf(rphivars_," %e %e",varr,varphi);
  } // End of loop over pairs
  fprintf(rphimeans_,"\n");
  fprintf(rphivars_,"\n");
  if(flush)
  {
    fflush(rphimeans_);
    fflush(rphivars_);
  }

  return;
}

// This function outputs 2D histograms of the magnitudes of complex fields
inline void rms_histograms_2d(char *ext_)
{
  static FILE *histogram_,*histogramtimes_; // histogram_ is an array of pointers to output files
  char name_[500];
  int i,j;
#if NDIMS>2
  int k;
#endif
  int fld1r=0,fld1i=1,fld2r=2,fld2i=3; // Indices of the fields
  int binnum[2]; // Indices of bins for a given pair of field values
  float binfreq[nbins0][nbins1]; // The frequency of field values occurring within each bin
  float bmin[2],bmax[2],df[2]; // Minimum and maximum field values for each field and spacing (in field values) between bins
  float magnitude1,magnitude2; // Magnitudes of the complex fields

  static int first=1;
  if(first) // Open output files
  {
    sprintf(name_,"rmshistogram2d%s",ext_);
    histogram_=fopen(name_,mode_);
    sprintf(name_,"rmshistogram2dtimes%s",ext_);
    histogramtimes_=fopen(name_,mode_);
    first=0;
  }

  fprintf(histogramtimes_,"%f",t); // Output time at which power spectra were recorded

  // Find the minimum and maximum values of the field - For now at least this is fixed
  bmin[0]=0.;
  bmax[0]=1.5;
  bmin[1]=0.;
  bmax[1]=1.5;

  // Find the difference (in field value) between successive bins
  df[0]=(bmax[0]-bmin[0])/(float)(nbins0);
  df[1]=(bmax[1]-bmin[1])/(float)(nbins1);
  
  // Initialize all frequencies to zero
  for(i=0;i<nbins0;i++)
    for(j=0;j<nbins1;j++)
      binfreq[i][j]=0.;

  // Iterate over grid to determine bin frequencies
  LOOP
  {
    magnitude1 = sqrt(pw2(FIELD(fld1r))+pw2(FIELD(fld1i)));; // Calculate magnitudes of fields
    magnitude2 = sqrt(pw2(FIELD(fld2r))+pw2(FIELD(fld2i)));
    binnum[0]=(int)((magnitude1-bmin[0])/df[0]); // Find index of bin for each value
    binnum[1]=(int)((magnitude2-bmin[1])/df[1]);
    if(magnitude1>=bmax[0]) // The maximal field value is at the top of the highest bin
      continue;
    if(magnitude1<=bmin[0])
      binnum[0]=0;
    if(magnitude2>=bmax[1])
      continue;
    if(magnitude2<=bmin[1])
      binnum[1]=0;
    if(!(binnum[0]>=0 && binnum[0]<nbins0) ||
       !(binnum[1]>=0 && binnum[1]<nbins1)) // This should never happen, but it doesn't hurt to have an extra check.
    {
      printf("Error in rmshistogram2d routine. Invalid bin number.\n");
      fprintf(histogram_,"Invalid bin number\n");
      exit(1);
    }
    binfreq[binnum[0]][binnum[1]]++;
  } // End of loop over grid

  // Output results
  for(i=0;i<nbins0;i++)
    for(j=0;j<nbins1;j++)
      fprintf(histogram_,"%e\n",binfreq[i][j]/(float)gridsize);
  fprintf(histogram_,"\n"); // Stick a blank line between times to make the file more readable
  fflush(histogram_);
  fprintf(histogramtimes_," %e %e %e %e",bmin[0],df[0],bmin[1],df[1]); // Output the starting point and stepsize for the bins at each time


  fprintf(histogramtimes_,"\n");
  fflush(histogramtimes_);
}

inline void model_output(int flush,char *ext_)
{
  if(sstrings && flush && t>=tstrings) // Find strings on the lattice
    find_strings(ext_);
  if(srmshistograms && flush) // Make histograms of rms values
    rms_histograms(ext_);
  if(srphimeansvars) // Output means and variances of amplitudes and phases
    rphimeansvars(ext_,flush);
  if(srmshistograms2d && flush) // Make 2d histograms of rms values
    rms_histograms_2d(ext_);
}

#undef PHIR
#undef PHII
#undef SIGMAR
#undef SIGMAI
#undef SIGMABARR
#undef SIGMABARI
#undef PHI2
#undef SIGMA2
#undef SIGMABAR2










