/*
MODEL.H file for the PLAINSYMMBREAK model.
Written by Gary Felder (gfelder@physics.stanford.edu)
Last Modified 3/7/01

This file contains the model specific functions and definitions for the model:
V = 1/4 lambda (phi^2 - v^2)^2
Comments: This model is for investigating symmetry breaking in a simple, one-field case. The field phi is assumed to begin at zero and roll down from there to its minimum at +-v.

The rescalings are set manually and expansion is neglected. So the constants beta, cpl, f0, r, and s have no effect.
The potential in program units is given by V_{pr} = 1/4 (phi_{pr}^2 - 1)^2
*/

/*
General comments about the model.h file:
This file contains the following functions - all called externally
modelinfo(FILE *info_) outputs information about the model and model-specific parameters to a file.
modelinitialize() performs any model-specific initialization
potential_energy(int term, float *field_values) calculates the average potential energy density term by term. The variable num_potential_terms (just above this function) specifies how many separate potential terms are used in this model.
dvdf(int fld, int i, int j, int k) calculates the potential term in the equation of motion, dV/dfield, for the field fld at the lattice point (i,j,k)
effective_mass(float mass_sq[], float *field_values) calculates the square masses of the fields and puts them in the array mass_sq. The parameter beginning tells the function to use initial field values - if this parameter is zero then the field quantities will be calculated dynamically.
model_output(int flush, char *ext_) allows each model to include its own specialized output function(s). The parameter flush is set to 1 when infrequent calculations are being performed and 0 otherwise. The string ext_ gives the extension for output filenames.
*/

/* - This section should be copied into parameters.h when using this model
// ---Adjustable parameters for PLAINSYMMBREAK model--- //
const float lambda=1.e-2; // Self-coupling of inflaton
const int sdomains=1; // Whether to output locations of domains
const float domain_cutoff=.5; // Cutoff used to define domains
*/

// Rescaling parameters.
// The program variables ("_pr") are defined as
//   f_pr = rescale_A a^rescale_r f (f=field value)
//   x_pr = rescale_B x (x=distance)
//   dt_pr = rescale_B a^rescale_s dt (t=time)
// The constants beta, cpl, and f0 are used to set the variable rescalings rescale_A, rescale_B, rescale_r, and rescale_s.
// These rescaling constants may be reset independently; their settings in terms of beta, cpl, and f0 are suggestions. See the documentation for more details.
// These rescalings are intrinsic to the model and probably shouldn't be changed for individual runs. Adjustable parameters are stored in "parameters.h"
const float beta=4.; // Exponent of the dominant term in the potential
const float cpl=lambda; // Coefficient of the dominant term in the potential (up to numerical factors - see documentation)
const float f0=1.; // Initial value of phi in Planck units, typically the point at which phi'=0
// By default these are automatically set to A=1/f0, B=sqrt(cpl) f0^(-1+beta/2), R=6/(2+beta), S=3(2-beta)/(2+beta). They may be adjusted to different values, but the relationship S=2R-3 must be maintained for the program equations to remain correct.
const float rescale_A=1.;
const float rescale_B=sqrt(lambda);
const float rescale_r=1.;
// The value of S in terms of R SHOULD NOT be changed.
const float rescale_s=2.*rescale_r-3.;

// Other global variables
// The array model_vars is intended to hold any model-specific, non-constant global variables.
// These variables should be initialized in modelinitialize() below
// Even if you're not using any, num_model_vars should be at least 1 to keep some compilers happy.
const int num_model_vars=1;
// Model specific variables: None are defined for this model.
extern float model_vars[num_model_vars];

// Macros to make the equations more readable: The only allowed value of fld is 0=Phi
#define PHI FIELD(0)

// Model specific details about the run to be output to an information file
inline void modelinfo(FILE *info_)
{
  // Name and description of model
  fprintf(info_,"Plain Symmetery Breaking Model\n");
  fprintf(info_,"V = 1/4 lambda (phi^2 - v^2)^2\n\n");

  // Model specific parameter values
  fprintf(info_,"lambda = %e\n",lambda);
}

// Perform any model specific initialization
// This function is called twice, once before initializing the fields (which_call=1) and once after (which_call=2)
inline void modelinitialize(int which_call)
{
  if(which_call==1)
  {
    if(nflds!=1)
    {
      printf("Number of fields for PLAINSYMMBREAK model must be 1. Exiting.\n");
      exit(1);
    }

    if(expansion!=0)
    {
      printf("Expansion is not implemented for PLAINSYMMBREAK model. Rerun with expansion=0.\n");
      exit(1);
    }
  }
}

// The constant num_potential_terms must be defined for use by outside functions
// Terms: term=0: 1/4 lambda (phi^2 - v^2)^2
const int num_potential_terms=1; // Number of terms that are calculated separately in the potential
// Potential energy terms
// See documentation for normalization of these terms.
// When setting initial conditions field values will be supplied in the array field_values. Otherwise this function will calculate them on the lattice.
inline float potential_energy(int term, float *field_values)
{
  DECLARE_INDICES
  float potential=0.;

  if(field_values==NULL) // If no values are given calculate averages on the lattice
  {
    // Loop over grid to calculate potential term
    LOOP
      potential += pw2(pw2(PHI)-1.);
    
    // Convert sum to average
    potential /= gridsize;
  }
  else // If field values are given then use them instead
    potential = pw2(pw2(field_values[0])-1.);

  // Include numerical coefficients
  potential *= .25;

  return (potential);
}

// Potential terms in the equations of motion, dV/dfield, evaluated at point (i,j,k)
// See documentation for details on the normalization of these terms
inline float dvdf(int fld, INDEXLIST)
{
  return( (pw2(PHI) - 1.)*PHI );
}

// Calculate effective mass squared and put it into the array mass_sq[] (used for initial conditions and power spectra)
// See documentation for normalization of these terms
// When setting initial conditions field values will be supplied in the array field_values. Otherwise this function will calculate them on the lattice.
inline void effective_mass(float mass_sq[], float *field_values)
{
  DECLARE_INDICES
  int fld;
  float fldsqrd[nflds]; // Square value of field
  float correction; // Used to adjust masses by the appropriate power of the scale factor. (See documentation.)
  
  // Loop over fields to find mean-square value
  if(field_values==NULL) // If no values are given calculate averages on the lattice
  {
    for(fld=0;fld<nflds;fld++)
    {
      fldsqrd[fld]=0.;
      LOOP
        fldsqrd[fld]+=pw2(FIELD(fld));
      fldsqrd[fld] /= (float)gridsize;
    }
  }
  else // For this model the initial modes are set with m=0 to simulate a quench
  {
    mass_sq[0]=0.;
    return;
  }
  
  mass_sq[0] = 3.*fldsqrd[0] - 1.;
    
  // Put in scale factor correction. This calculation should be the same for all models.
  if(expansion>0) // If there's no expansion don't bother with this.
  {
    correction = pow(a,2.*rescale_s+2.);
    for(fld=0;fld<nflds;fld++)
      mass_sq[fld] *= correction;
  }
}

////////////////////////////////////////////////////////////////////////////////////
// Model-specific output functions - The function model_output is called externally/
////////////////////////////////////////////////////////////////////////////////////

// This function finds all points on the lattice in "strings," i.e. with magnitude(sigma) less than a preset cutoff
inline void find_domains(char *ext_)
{
  static FILE *plusdomains_,*minusdomains_,*domaintimes_;
  char name_[500]; // Used for filenames
  int i,j,k;
  int numptsplus=0,numptsminus=0; // Number of points in positive and negative domains at a given time

  static int first=1;
  if(first) // Open output files
  {
    sprintf(name_,"plusdomains%s",ext_);
    plusdomains_=fopen(name_,mode_);
    sprintf(name_,"minusdomains%s",ext_);
    minusdomains_=fopen(name_,mode_);
    sprintf(name_,"domaintimes%s",ext_);
    domaintimes_=fopen(name_,mode_);
    first=0;
  }

  LOOP
  {
    if(PHI>domain_cutoff)
    {
      if(NDIMS==1)
        fprintf(plusdomains_,"%d\n",i);
      else if(NDIMS==2)
        fprintf(plusdomains_,"%d %d\n",i,j);
      else if(NDIMS==3)
        fprintf(plusdomains_,"%d %d %d\n",i,j,k);
      numptsplus++;
    }
    else if(PHI<-domain_cutoff)
    {
      if(NDIMS==1)
        fprintf(minusdomains_,"%d\n",i);
      else if(NDIMS==2)
        fprintf(minusdomains_,"%d %d\n",i,j);
      else if(NDIMS==3)
        fprintf(minusdomains_,"%d %d %d\n",i,j,k);
      numptsminus++;
    }
  }

  fprintf(domaintimes_,"%f %d %d\n",t,numptsplus,numptsminus);
  fflush(plusdomains_);
  fflush(minusdomains_);
  fflush(domaintimes_);
  
  return;
}

inline void model_output(int flush,char *ext_)
{
  if(sdomains && flush) // Find domains on the lattice
    find_domains(ext_);
}

#undef PHI










