/*
MODEL.H file for the QINFLATION model.
Written by Gary Felder (gfelder@physics.stanford.edu)
Last Modified 3/7/01

This file contains the model specific functions and definitions for
the model:
V = V_0 exp(-2 v/|phi|) + 1/4 lambda |phi|^4
Comments: The field phi is complex, |phi|^2 = phi_R^2 + phi_I^2. This
model is currently only implemented for use without expansion.

We define new parameters vpr = 4 pi^(1/4) sqrt(v) and lpr =
vpr^3/(2^12 pi^2 V_0) lambda. We do not use the default rescalings
(see below), but rather set A=8 sqrt(pi)/vpr, B=8 sqrt(pi V_0/vpr).

The potential in program units is given by V_{pr} = 1/vpr
exp(-vpr/|phi_pr|) + 1/4 lpr |phi_pr|^4.  */

/*
General comments about the model.h file:
This file contains the following functions - all called externally
modelinfo(FILE *info_) outputs information about the model and model-specific parameters to a file.
modelinitialize() performs any model-specific initialization
potential_energy(int term, float *field_values) calculates the average potential energy density term by term. The variable num_potential_terms (just above this function) specifies how many separate potential terms are used in this model.
dvdf(int fld, int i, int j, int k) calculates the potential term in the equation of motion, dV/dfield, for the field fld at the lattice point (i,j,k)
effective_mass(float mass_sq[], float *field_values) calculates the square masses of the fields and puts them in the array mass_sq. The parameter beginning tells the function to use initial field values - if this parameter is zero then the field quantities will be calculated dynamically.
model_output(int flush, char *ext_) allows each model to include its own specialized output function(s). The parameter flush is set to 1 when infrequent calculations are being performed and 0 otherwise. The string ext_ gives the extension for output filenames.
*/

/* - This section should be copied into parameters.h when using this model
// ---Adjustable parameters for QINFLATION model--- //
const float alpha=200.; // The free parameter in the model (see paper)
const float V0=5.93e-13/alpha; // Characteristic energy density
const float vpr=1.57*pow(alpha,-.75); // Characteristic field value.
const float lpr=.01; // Rescaled self-coupling of phi
const int sqballs=0; // Whether to find qballs
const float qball_cutoff=.5*vpr; // Cutoff value for what is considered a q-ball
const float tqballs=0.; // Time at which to start looking for q-balls
const int sqhist=1; // Whether to calculate a q histogram
*/

// Rescaling parameters.
// The program variables ("_pr") are defined as
//   f_pr = rescale_A a^rescale_r f (f=field value)
//   x_pr = rescale_B x (x=distance)
//   dt_pr = rescale_B a^rescale_s dt (t=time)
// For this model A and B are not set using the defaults. The rescalings r and s are irrelevant without expansion.
const float rescale_A=2.*sqrt(pi)/vpr;
const float rescale_B=2.*sqrt(pi*V0/vpr);
const float rescale_r=0.;
// The value of S in terms of R SHOULD NOT be changed.
const float rescale_s=2.*rescale_r-3.;

// Other global variables
// The array model_vars is intended to hold any model-specific, non-constant global variables.
// These variables should be initialized in modelinitialize() below
// Even if you're not using any, num_model_vars should be at least 1 to keep some compilers happy.
const int num_model_vars=1;
// Model specific variables: None are defined for this model.
extern float model_vars[num_model_vars];

// Macros to make the equations more readable: The values of fld are 0=Phi,1=Sigma
#define PHIR FIELD(0)
#define PHII FIELD(1)
#define PHI2 (pw2(PHIR)+pw2(PHII))
#define PHI (sqrt(PHI2))
#define Q (FIELD(0)*FIELDD(1)-FIELD(1)*FIELDD(0))

// Model specific details about the run to be output to an information file
inline void modelinfo(FILE *info_)
{
  // Name and description of model
  fprintf(info_,"Q Inflation Model\n");
  fprintf(info_,"V = V_0 exp(-2 v/|phi|) + 1/4 lambda |phi|^4\n\n");

  // Model specific parameter values
  fprintf(info_,"alpha = %e\n",alpha);
  fprintf(info_,"V_0 = %e\n",V0);
  fprintf(info_,"vpr = %e\n",vpr);
  fprintf(info_,"lpr = %e\n",lpr);
}

// Perform any model specific initialization
// This function is called twice, once before initializing the fields (which_call=1) and once after (which_call=2)
inline void modelinitialize(int which_call)
{
  if(which_call==1)
  {
    if(nflds!=2)
    {
      printf("Number of fields for QINFLATION model must be 2. Exiting.\n");
      exit(1);
    }
    if(expansion!=0)
    {
      printf("QINFLATION model has not been implemented with expansion. To run  this model reset expansion to zero.\n");
      exit(1);
    }
  }
}

// The constant num_potential_terms must be defined for use by outside functions
// Terms: term=0: V_0 exp(-2 v/|phi|) -- term=1: 1/4 lambda |phi|^4
const int num_potential_terms=2; // Number of terms that are calculated separately in the potential
// Potential energy terms
// See documentation for normalization of these terms.
// When setting initial conditions field values will be supplied in the array field_values. Otherwise this function will calculate them on the lattice.
inline float potential_energy(int term, float *field_values)
{
  DECLARE_INDICES
  float potential=0.;

  if(field_values==NULL) // If no values are given calculate averages on the lattice
  {
    // Loop over grid to calculate potential term
    LOOP
    {
      if(term==0)
        potential += exp(-vpr/PHI);
      if(term==1)
        potential += pw2(PHI2);
    }

    // Convert sum to average
    potential /= gridsize;
  }
  else // If field values are given then use them instead
  {
    if(term==0)
      potential = exp(-vpr/sqrt(pw2(field_values[0])+pw2(field_values[1])));
    if(term==1)
      potential = pw2(pw2(field_values[0])+pw2(field_values[1]));
  }

  // Include numerical coefficients
  if(term==0)
    potential /= vpr;
  if(term==1)
    potential *= .25*lpr;

  return (potential);
}

// Potential terms in the equations of motion, dV/dfield, evaluated at point (i,j,k)
// See documentation for details on the normalization of these terms
inline float dvdf(int fld, INDEXLIST)
{
  // Note: This is incredibly inefficient. Some of these redundant calculations could be avoided.
  if(fld==0) // Phi_R
    return( (exp(-vpr/PHI)/PHI2/PHI + lpr*PHI2)*PHIR );
  else // Phi_I
    return( (exp(-vpr/PHI)/PHI2/PHI + lpr*PHI2)*PHII );
}

// Calculate effective mass squared and put it into the array mass_sq[] (used for initial conditions and power spectra)
// See documentation for normalization of these terms
// When setting initial conditions field values will be supplied in the array field_values. Otherwise this function will calculate them on the lattice.
inline void effective_mass(float mass_sq[], float *field_values)
{
  DECLARE_INDICES
  int fld;
  float correction; // Used to adjust masses by the appropriate power of the scale factor. (See documentation.)
  float phi,ephi,phi3,combo;

  // Loop over fields to find mean-square value
  if(field_values==NULL) // If no values are given calculate averages on the lattice
  {
    mass_sq[0]=0.;
    mass_sq[1]=0.;
    LOOP
    {
      phi=PHI; // Prevents recalculating this each time it's used
      ephi=exp(-vpr/phi);
      phi3=1./phi/phi/phi;
      combo=vpr*pw2(phi3)-3.*phi3/pw2(phi);
      // This is very inefficient as it is. Among other things I should pull coefficients like lpr out of the loop.
      mass_sq[0] += (phi3+combo*pw2(PHIR))*ephi + lpr*(3.*pw2(PHIR)+pw2(PHII));
      mass_sq[1] += (phi3+combo*pw2(PHII))*ephi + lpr*(3.*pw2(PHII)+pw2(PHIR));
    }
    mass_sq[0] /= gridsize;
    mass_sq[1] /= gridsize;
  }
  else // If field values are given then use them instead
  {
    phi=sqrt(pw2(field_values[0])+pw2(field_values[1]));
    ephi=exp(-vpr/phi);
    phi3=1./phi/phi/phi;
    combo=vpr*pw2(phi3)-3.*phi3/pw2(phi);
    mass_sq[0] = (phi3+combo*pw2(field_values[0]))*ephi + lpr*(3.*pw2(field_values[0])+pw2(field_values[1]));
    mass_sq[1] = (phi3+combo*pw2(field_values[1]))*ephi + lpr*(3.*pw2(field_values[1])+pw2(field_values[0]));
  }

  for(fld=0;fld<nflds;fld++) // Don't allow negative m^2 as this will cause numerical errors.
    if(mass_sq[fld]<0.)
      mass_sq[fld]=0.;

  // Put in scale factor correction. This calculation should be the same for all models.
  if(expansion>0) // If there's no expansion don't bother with this.
  {
    correction = pow(a,2.*rescale_s+2.);
    for(fld=0;fld<nflds;fld++)
      mass_sq[fld] *= correction;
  }
}

////////////////////////////////////////////////////////////////////////////////////
// Model-specific output functions - The function model_output is called externally/
////////////////////////////////////////////////////////////////////////////////////

// This function finds all points on the lattice with Q (defined at the top of file) greater than a preset cutoff
inline void find_qballs(char *ext_)
{
  static FILE *qballs_,*negqballs_,*qtimes_;
  char name_[500]; // Used for filenames
  int i,j,k;
  int numqpts=0,numnegqpts=0; // Number of points in q-balls and negative q-balls at a given time

  static int first=1;
  if(first) // Open output files
  {
    sprintf(name_,"qballs%s",ext_);
    qballs_=fopen(name_,mode_);
    sprintf(name_,"negqballs%s",ext_);
    negqballs_=fopen(name_,mode_);
    sprintf(name_,"qtimes%s",ext_);
    qtimes_=fopen(name_,mode_);
    first=0;
  }

  LOOP
  {
    if(Q>qball_cutoff) // All values above the cutoff are considered part of a q-ball
    {
      if(NDIMS==1)
        fprintf(qballs_,"%d\n",i);
      else if(NDIMS==2)
        fprintf(qballs_,"%d %d\n",i,j);
      else if(NDIMS==3)
        fprintf(qballs_,"%d %d %d\n",i,j,k);
      numqpts++;
    }
    if(Q<-qball_cutoff) // Negative values larger than the cutoff are "negative q-balls"
    {
      if(NDIMS==1)
        fprintf(negqballs_,"%d\n",i);
      else if(NDIMS==2)
        fprintf(negqballs_,"%d %d\n",i,j);
      else if(NDIMS==3)
        fprintf(negqballs_,"%d %d %d\n",i,j,k);
      numnegqpts++;
    }
  }
  fprintf(qtimes_,"%f %d %d\n",t,numqpts,numnegqpts);
  fflush(qballs_);
  fflush(negqballs_);
  fflush(qtimes_);
  
  return;
}

// This function is identical to the generic histogram function in output.cpp except this makes a histogram of the q charge, phi_R phi_I' - phi_I phi_R'
// It uses the same number of bins specified for the normal histograms (nbins)
inline void qhist(char *ext_)
{
  static FILE *qhist_,*qhisttimes_;
  char name_[500]; // Used for filenames
  int i,j,k;
  int binnum; // Index of bin for a given field value
  float binfreq[nbins]; // The frequency of field values occurring within each bin
  float bmin,bmax,df; // Minimum and maximum field values for each field and spacing (in field values) between bins
  float q; // Q charge density. Using this variable avoids recalculating it several times at each point.

  static int first=1;
  if(first) // Open output files
  {
    sprintf(name_,"qhist%s",ext_);
    qhist_=fopen(name_,mode_);
    sprintf(name_,"qhisttimes%s",ext_);
    qhisttimes_=fopen(name_,mode_);
    first=0;
  }

  fprintf(qhisttimes_,"%f",t); // Output time at which power spectra were recorded
  
  // Find the minimum and maximum values of the field
  i=0;j=0;k=0;
  bmin=Q;
  bmax=bmin;
  LOOP
  {
    q=Q; // Calculate Q charge density
    bmin = (q<bmin ? q : bmin);
    bmax = (q>bmax ? q : bmax);
  }
  
  // Find the difference (in field value) between successive bins
  df=(bmax-bmin)/(float)(nbins); // bmin will be at the bottom of the first bin and bmax at the top of the last
  
  // Initialize all frequencies to zero
  for(i=0;i<nbins;i++)
    binfreq[i]=0.;
  
  // Iterate over grid to determine bin frequencies
  LOOP
  {
    q=Q; // Calculate Q charge density
    binnum=(int)((q-bmin)/df); // Find index of bin for each value
    if(q>=bmax) // The maximal field value is at the top of the highest bin
      binnum=nbins-1;
    if(binnum>=0 && binnum<nbins) // Increment frequency in the appropriate bin
      binfreq[binnum]++;
    else // This should never happen, but it doesn't hurt to have an extra check.
    {
      printf("Error in qhist routine. Invalid bin number.\n");
      fprintf(qhist_,"Invalid bin number\n");
    }
  } // End of loop over grid
  
  // Output results
  for(i=0;i<nbins;i++)
    fprintf(qhist_,"%e\n",binfreq[i]/(float)gridsize); // Output bin frequency normalized so the total equals 1
  fprintf(qhist_,"\n"); // Stick a blank line between times to make the file more readable
  fflush(qhist_);
  fprintf(qhisttimes_," %e %e",bmin,df); // Output the starting point and stepsize for the bins at each time
  fprintf(qhisttimes_,"\n");
  fflush(qhisttimes_);
}

inline void model_output(int flush,char *ext_)
{
  if(sqballs && flush && t>=tqballs) // Find qballs on the lattice
    find_qballs(ext_);
  if(sqhist && flush) // Do a histogram of q charge
    qhist(ext_);
}

#undef PHIR
#undef PHII
#undef PHI2
#undef PHI
